# FMV Shot Logger
# shotmanager
#
# Logs data when the camera trigger starts/stops to sync with video
#
# Created by Phu Nguyen 6/10/2016
#  Copyright (c) 2016 3D Robotics. All rights reserved.


#!/usr/bin/env python

import os
import time
import math

import shotLogger
from GoProConstants import *
import shots
import app_packet
import struct

logger = shotLogger.logger
LOG_DIR = '/log/fmv/'

GOPRO_WIDE_FOV = 0
GOPRO_MEDIUM_FOV = 1
GOPRO_NARROW_FOV = 2

class FMVLogger():
	
	def __init__(self, shotmgr):
		# initialize shotmanager object
		self.shotmgr = shotmgr
		self.vehicle = self.shotmgr.vehicle
		self.dataFile = None

		self.registerCallbacks()
		# Update the gopro time
		self.shotmgr.goproManager.update_gopro_time(time.time())

	def setupDataFile(self):
		if not os.path.exists(LOG_DIR):
			os.makedirs(LOG_DIR)

		unique_id = time.strftime('%Y_%m_%d_%H_%M_%S')
		file_name = 'fmvLog_{0}.txt'.format(unique_id)
		file_path = LOG_DIR + file_name

		if self.dataFile is None:
			try:
				logger.log('[FMVLogger]: Opening shot data file @ %s.' % file_path)
				self.dataFile = open(file_path, 'ab')
			except IOError:
				logger.log('[FMVLogger]: Unable to open shot data file @ %s.' % file_path)
				self.dataFile = None
			
						
	def vehicleStateCallback(self, vehicle, property, value):
		if self.shotmgr.goproManager.isRecording == False:
			return
		if self.dataFile is None:
			return
		# Calls back everytime attitude of the vehicle changes
		current_time = time.time()
		#self.shotmgr.shotDataFile.write('Vehicle: %d, %d, %d', self.vehicle.attitude.pitch, self.vehicle.attitude.roll, self.vehicle.attitude.yaw)
		self.dataFile.write('{0},'.format(current_time*1000000)) #Unix Time Stamp in Microseconds FMV Multiplexer needs Microseconds
		self.dataFile.write('{0},{1},{2},'.format(vehicle.heading, math.degrees(vehicle.attitude.pitch), math.degrees(vehicle.attitude.roll))) #Convert radians in pitch, roll, yaw into degrees
		if vehicle.location.global_relative_frame is not None:
			self.dataFile.write('{0},{1},{2},'.format(vehicle.location.global_relative_frame.lat, vehicle.location.global_relative_frame.lon, vehicle.location.global_relative_frame.alt))
		if self.vehicle.gimbal.yaw is not None:
			fov = self.horizontalFOVLookup(self.shotmgr.goproManager.videoFieldOfView)
			self.dataFile.write('{0},{1},{2},{3}'.format(fov, self.vehicle.gimbal.yaw, vehicle.gimbal.pitch - math.degrees(vehicle.attitude.pitch), math.degrees(vehicle.attitude.roll))) #Roll Angle Relative  = Roll Angle Platform, Pitch Relative angle sensor - angle platform

		self.dataFile.write('\n')
		self.dataFile.flush()

	def startLogging(self):
		logger.log('[FMVLogger] Started Logging')
		self.setupDataFile()
		if self.dataFile is not None:
			self.dataFile.write('UnixTimeStamp,PlatformHeading,PlatformPitch, PlatformRoll, SensorLatitude,SensorLongitude,SensorAltitude,HorizontalFOV, SensorRelativeAzimuth, SensorRelativeElevation, SensorRelativeRoll \n')
			self.dataFile.flush()

	def stopLogging(self):
		logger.log('[FMVLogger] STOPPED Logging')
		self.dataFile = None

	def handlePacket(self, packetType, packetValue):
		try:
			if packetType == app_packet.GOPRO_RECORD:
				(startstop,) = struct.unpack('<I', packetValue)
				self.toggleLogging(startstop)
				
		except Exception as e:
			logger.log('[FMVLogger] Error handling packet (%s)' % e)

	def toggleLogging(self, command):
		if self.shotmgr.goproManager.captureMode == CAPTURE_MODE_VIDEO:
			if command == RECORD_COMMAND_TOGGLE:
				if self.shotmgr.goproManager.isRecording:
					self.stopLogging()
				else:
					self.startLogging()
			elif command == RECORD_COMMAND_START:
				self.startLogging()
			else:
				self.stopLogging()

	def registerCallbacks(self):
		self.vehicle.add_attribute_listener('attitude', self.vehicleStateCallback)

	def horizontalFOVLookup(self, fov):
		if fov == GOPRO_WIDE_FOV:
			return 118.2
		elif fov == GOPRO_MEDIUM_FOV:
			return 94.4
		else:
			return 64.4
